/*
Copyright (C) 2013-2018 Draios Inc dba Sysdig.

This file is part of sysdig.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

*/

#define SE_NSIG           64
#define SE_SIGHUP           1
#define SE_SIGINT           2
#define SE_SIGQUIT          3
#define SE_SIGILL           4
#define SE_SIGTRAP          5
#define SE_SIGABRT          6
#define SE_SIGIOT           6
#define SE_SIGBUS           7
#define SE_SIGFPE           8
#define SE_SIGKILL          9
#define SE_SIGUSR1         10
#define SE_SIGSEGV         11
#define SE_SIGUSR2         12
#define SE_SIGPIPE         13
#define SE_SIGALRM         14
#define SE_SIGTERM         15
#define SE_SIGSTKFLT       16
#define SE_SIGCHLD         17
#define SE_SIGCONT         18
#define SE_SIGSTOP         19
#define SE_SIGTSTP         20
#define SE_SIGTTIN         21
#define SE_SIGTTOU         22
#define SE_SIGURG          23
#define SE_SIGXCPU         24
#define SE_SIGXFSZ         25
#define SE_SIGVTALRM       26
#define SE_SIGPROF         27
#define SE_SIGWINCH        28
#define SE_SIGIO           29
#define SE_SIGPOLL         SE_SIGIO
#define SE_SIGPWR          30
#define SE_SIGSYS          31
#define SE_SIGUNUSED       31
