LIBRARY scap

EXPORTS
		scap_open_live
		scap_open_offline
		scap_open_offline_fd
		scap_open
		scap_close
		scap_get_os_platform
		scap_get_ndevs
		scap_getlasterr
		scap_next
		scap_event_getlen
		scap_event_get_ts
		scap_dump_open
		scap_dump_open_fd
		scap_dump_close
		scap_dump_get_offset
		scap_dump_flush
		scap_dump_ftell
		scap_dump
		scap_event_reset_count
		scap_event_get_num
		scap_get_proc_table
		scap_event_getinfo
		scap_stop_capture
		scap_get_ifaddr_list
		scap_get_stats
		scap_get_event_info_table
		scap_get_syscall_info_table
		scap_proc_get
		scap_proc_free
		scap_start_capture
		scap_get_machine_info
		scap_stop_dropping_mode
		scap_start_dropping_mode
		scap_get_user_list
		scap_free_userlist
		scap_set_snaplen
		scap_get_readfile_offset
		scap_clear_eventmask
		scap_set_eventmask
		scap_unset_eventmask
		scap_number_of_bytes_to_write
		scap_event_get_dump_flags
		scap_enable_dynamic_snaplen
		scap_disable_dynamic_snaplen
		scap_proc_free_table
		scap_is_thread_alive
		scap_get_threadlist
		scap_get_host_root
		scap_ftell
		scap_fseek
